/**
 * The bridge to use Ogre::DataStream with the C++ Streams.
 *
 * The OgreStream class is released under the public domain, whereas
 * the OgreStreambuf class is based on the work of Edd, so it's released
 * under the Boost Software License - Version 1.0 (see the next comment).
 */

/*

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

*/

#include <streambuf>
#include <vector>
#include <cstdlib>
#include <algorithm>
#include <cstring>
#include <stdexcept>

#include "OgreDataStream.h"

/**
 * The stream buffer to integrate the stream with Ogre::DataStream.
 *
 * This class is based on the work of Edd, available at
 *   http://www.mr-edd.co.uk/blog/beginners_guide_streambuf .
 *
 * This class is released under the Boost Software License - Version 1.0.
 * Please see at the beginning of the file for the full text.
 */
class OgreStreambuf : public std::streambuf {
public:
	OgreStreambuf(Ogre::DataStreamPtr datastream, std::size_t buffSize = 256,
			std::size_t putBack = 8);

protected:
    int_type underflow();
    std::streamsize showmanyc();

    std::streampos seekoff(std::streamoff off, std::ios_base::seekdir way,
    		std::ios_base::openmode which = std::ios_base::in);
    std::streampos seekpos(std::streampos sp,
    		std::ios_base::openmode which = std::ios_base::in);


private:
	Ogre::DataStreamPtr mDatastream;
	const std::size_t mPutBack;
	std::vector<char> mBuffer;
};

/**
 * The stream that uses Ogre::DataStream.
 * Actually it's a shortcut, because it just has a modified callback in order
 * to be quicker when using the stream.
 */
class OgreStream : public std::istream {
public:
	OgreStream(Ogre::DataStreamPtr datastream, std::size_t buffSize = 256,
			std::size_t putBack = 8);
	~OgreStream();
};

/**
 * The constructor of the stream buffer.
 * Initializes the members.
 */
OgreStreambuf::OgreStreambuf(Ogre::DataStreamPtr datastream,
		std::size_t buffSize, std::size_t putBack) :
		mDatastream(datastream),
		mPutBack(std::max(putBack, size_t(1))),
		mBuffer(std::max(buffSize, putBack))
{

	char *end = &mBuffer.front() + mBuffer.size();
	setg(end, end, end);
}

/**
 * Read from the datastream.
 */
std::streambuf::int_type OgreStreambuf::underflow()
{
	// buffer not exhausted
	if(gptr() < egptr()) {
        return traits_type::to_int_type(*gptr());
	}

	if(mDatastream->eof()) {
		return traits_type::eof();
	}

	char *base = &mBuffer.front();
	char *start = base;

	if(eback() == base) {
		// Make arrangements for putback characters
		std::memmove(base, egptr() - mPutBack, mPutBack);
		start += mPutBack;
    }

    // Read from fptr_ in to the provided buffer
	std::size_t read = mDatastream->read(start,
			mBuffer.size() - (start - base));

	// Set buffer pointers
	setg(base, start, start + read);

	return traits_type::to_int_type(*gptr());
}

/**
 * Tells how many characters can be read.
 */
std::streamsize OgreStreambuf::showmanyc()
{
	return mDatastream->size() - mDatastream->tell();
}

/**
 * Seek the position in the file.
 */
std::streampos OgreStreambuf::seekoff(std::streamoff off,
		std::ios_base::seekdir way, std::ios_base::openmode which)
{
	using namespace std;

	/*
	 * There isn't anything about exception in datastream documentation,
	 * however there can be user defined datastreams, so use the try-catch.
	 */
	try {
		switch(way) {
		case ios_base::cur:
			mDatastream->skip(off);
			break;

		case ios_base::end:
			off = mDatastream->size() - off - 1;
			// no break, since we corrected the offset to be from the beginning

		case ios_base::beg:
			mDatastream->seek(off);
			break;

		default:
			// Unknown? Return -1!
			return -1;
		}
	} catch(std::exception &e) {
		return -1;
	}

	return static_cast<std::streampos>(mDatastream->tell());
}

/**
 * Set the absolute position in the file.
 */
std::streampos OgreStreambuf::seekpos(std::streampos ps,
		std::ios_base::openmode which)
{
	return seekoff(ps, std::ios_base::beg, which);
}

/**
 * The constructor of the stream.
 * Creates a new OgreStreambuf and pass to istream.
 */
OgreStream::OgreStream(Ogre::DataStreamPtr datastream,
		std::size_t buffSize, std::size_t putBack) :
		std::istream(new OgreStreambuf(datastream, buffSize, putBack))
{
}

/**
 * The destructor of the stream, deletes the OgreStreambuf.
 */
OgreStream::~OgreStream()
{
	if(typeid(*rdbuf()) == typeid(OgreStreambuf)) {
		delete rdbuf();
	}
}
